% Input (squared) symmetric ssimilarities Data (can be a pointer)
% SamplingIndices
% N - number of points accessible in Data
classdef NystroemContainer < DataContainer
    properties (GetAccess='public',SetAccess = 'private')
    end
    
    properties (GetAccess='private', SetAccess = 'private')
        SamplingIndices = [];
        Smm             = [];
        Snm             = [];
        PInvSmm         = [];
    end
    
    
    methods
        % if constructor is called with a cell is expected to be given from GetParameters-Cell 
        function obj = NystroemContainer(Data,SamplingIndices,N)
            if(iscell(Data))
                SamplingIndices = Data{2};
                N               = size(Data{1},1);
            end
            
            if(~iscell(Data))
                if(numel(SamplingIndices)>2000 || numel(SamplingIndices)>N)
                    warning('Sampling to large - define own sampling with upto 500 points');
                    R = randperm(N);
                    SamplingIndices = R(1:min(500,N));
                end
            end
            
            obj@DataContainer([]); 
            if(iscell(Data))
                obj.SamplingIndices = SamplingIndices;
                obj.N               = N;            
                obj.M               = N;            
                obj.Snm = Data{1}(1:N,:);
                obj.Smm = double(obj.Snm(SamplingIndices,:));
                obj.Smm = 0.5 * (obj.Smm+obj.Smm');
                obj.Snm(SamplingIndices,:) = obj.Smm;
                obj.PInvSmm = pinv(full(obj.Smm));    
            else
                obj.SamplingIndices = SamplingIndices;
                obj.N               = N;            
                obj.M               = N;            
                obj.Snm = Data(1:N,SamplingIndices);
                obj.Smm = double(obj.Snm(SamplingIndices,:));
                obj.Smm = 0.5 * (obj.Smm+obj.Smm');
                obj.Snm(SamplingIndices,:) = obj.Smm;
                obj.PInvSmm = pinv(full(obj.Smm));    
            end
        end

        % returns the parameter (not data matrix) of the binded container
        function oCell = GetParameter(me,varargin)
            if(~isempty(varargin))
                switch varargin{1}
                    case 1
                        oCell = me.Snm;
                    case 2
                        oCell = me.SamplingIndices;
                    case 3
                        oCell = me.PInvSmm;
                end
            else
                oCell = {me.Snm,me.SamplingIndices,me.PInvSmm};
            end
        end
        
        
        function ret = mtimes(A,B)
            X = [];
            bLeft = false;
            if(isnumeric(A))
                me = B;
                X  = A;
                bLeft = true;
            else
                me = A;
                X  = B;
            end
            % vector times nystroem approximated matrix
            % this way the full matrix calculation is avoided
            if(~isvector(X))
                if(~bLeft)
                    M = size(X,2); % 
                    [m_N,m_M]=size(me);
                    ret = zeros(M,m_N);
                    for(k=1:M)
                        ret(k,:) = me.Snm*(X(:,k)'*me.Snm * me.PInvSmm)';
                    end
                else
                    M = size(X,1); % 
                    [m_N,m_M]=size(me);
                    ret = zeros(M,m_N);
                    for(k=1:M)
                        ret(k,:) = me.Snm*(X(k,:)*me.Snm * me.PInvSmm)';
                    end                    
                end
            else
                ret = me.Snm*(X'*me.Snm * me.PInvSmm)';
            end
        end
            
        function Value = subsref(me,Indices)
            Value = [];
            NumInd = numel(Indices);
            if(NumInd==1) % direct access 
                Subs       = Indices.subs;
                NumIndSubs = numel(Subs);
                switch Indices.type
                    case '()'     % indexing                        
                        if(NumIndSubs==2)
                            Value = me.CalculateValue(Subs{1},Subs{2});
                        else
                            ME = MException('Class:NystroemContainer:Error','number of parameters must be two');
                            throw(ME);
                        end
                    case '{}'
                        if(NumIndSubs==1)
                            Value = me.GetParameter(Subs{1});
                        else
                            ME = MException('Class:NystroemContainer:Error','number of cell parameters must be one');
                            throw(ME);
                        end
                    case '.'     % class member
                        Value = me.(Subs);
                end
            else % should be a function call
                if(strncmp(Indices(1).type,'.',1))   
                    Value = builtin('subsref', me, Indices); 
                end
            end
        end       
        
        function ret = GetData(me,X,Y)            
            if(~exist('X','var'))
                ret = me.Snm*me.PInvSmm*me.Snm';
            else
                ret = me.Snm(X,:)*me.PInvSmm*me.Snm(Y,:)';
            end
        end
        
        function ret = GetDiags(me)
                N               = size(me.Snm,1);
                iBlockSize      = 1000;
                ret             = zeros(N,1);
                for(k=1:iBlockSize:N)  % go through all sequences
                    vIndices = k:min(k+iBlockSize-1,N);
                    ret(vIndices)=sum( (me.PInvSmm*me.Snm(vIndices,:)') .* me.Snm(vIndices,:)');
                end
        end
        
        % input 
        % X - index vector 
        % Y - index vector 
        function Value = CalculateValue(me,X,Y)       
            X = X(:);
            Y = Y(:);   
            Value = me.GetData(X,Y);
        end
    end
        
     methods(Access = protected)
       % Override copyElement method:
      function cpObj = copyElement(obj)
         % Make a shallow copy of all four properties
         cpObj = copyElement@matlab.mixin.Copyable(obj);      
      end
   end
end