% Indefinite core vector machine - with linear runtime but output model is
% not sparse 
function [oModelCore,fClassifierCore,dAcc] = IndefiniteCoreVectorMachine(oNystroemDataContainer,C,Y,dEpsilon)
if(~exist('dEpsilon','var'))
    dEpsilon = 0.00000001; % should be sufficiently small
end
N          = numel(Y);
sInfo      = whos('oNystroemDataContainer');

if(ismatrix(oNystroemDataContainer) || strcmp(sInfo.class,'DataContainerFilter'))        
    if(N>1000) 
        % if we have many points we have to use a Nystroem sampling or other to get the rest efficient
        % the following is a native sampling approach and could be improved      
        v0 = find(Y==-1); N0 = numel(v0);
        v1 = find(Y==1);  N1 = numel(v1);
        vLandmarks0 = randsample(N0,min(N0,min(ceil(N0*0.30),100)));
        vLandmarks1 = randsample(N1,min(N1,min(ceil(N1*0.30),100)));
        vLandmarks = [v0(vLandmarks0); v1(vLandmarks1)];
        %vLandmarks = randsample(N,min(N,min(ceil(N*0.3),200)));
    else
        % do not really approximate - all samples used 
        vLandmarks = 1:N;
    end
	Kext       = oNystroemDataContainer;
else
    Kext       = oNystroemDataContainer.GetData; % only for acc
    vLandmarks = oNystroemDataContainer.GetParameter{2};
end

% get a Nystroem approximated matrix 
% needed to have the eigendecomposition cheap
mTemp      = zeros(N,numel(vLandmarks));
mTemp(vLandmarks,:) = diag(diag(ones(numel(vLandmarks))))/C;        
ny_K{1}     = oNystroemDataContainer(:,vLandmarks) .* bsxfun(@times,Y,Y(vLandmarks)') +  bsxfun(@times,Y,Y(vLandmarks)') + mTemp;
clear mTemp;
ny_K{2}     = pinv(ny_K{1}(vLandmarks,:));
[e, v, idx] = eig_ny(ny_K);
clear ny_K;
V           = e;
clear       e;
vNeg        = sign(diag(v))<0; 

% create a Nystroem approximated modified kernel
% by using indefinite SVM approach of Loosli
'trust on weak evaluation of K in meb function'
CW          = V * (abs(v));
KTilde      = @(X,Y)CW(X,:)*V(Y,:)';
vDiags      = zeros(N,1); 

% calculate diagonal elements (can be
% skipped if diagonal values of the kernel are known 
if(N>=4000)
    ny_Kabs{1}     = CW * V(vLandmarks,:)';
    clear CW;    
    ny_Kabs{2}     = pinv(ny_Kabs{1}(vLandmarks,:));
    vDiags         = sum((ny_Kabs{1} * ny_Kabs{2}') .* ny_Kabs{1},2); % approximated diagonal elements    
    clear ny_Kabs;
end

if(N<4000)
    KTilde = KTilde(1:N,1:N);
    vDiags = diag(KTilde);
end

% calculate the acutal core vector machine solution on
% the modified kernel (modification used indefinite SVM approach
% of Loosli) adapted for CVM
Delta              = (max(vDiags)-vDiags);
Kappa              = max(vDiags);        
[S,R,vAlpha]       = core_set_one_two(KTilde,dEpsilon,vDiags,Delta,Kappa); % using only the kernel
clear CW;
clear KTilde;
oModelCore.R           = R.^2;
oModelCore.CoreVectors = S;
oModelCore.Alphas      = vAlpha;

% get final model right by projecting the solution vector
% in the Krein space see Loosli et al.
oModelCore.Alphas      = V(oModelCore.CoreVectors,:)'*oModelCore.Alphas;
oModelCore.Alphas(vNeg)= -oModelCore.Alphas(vNeg);
oModelCore.Alphas      = (V*oModelCore.Alphas);

oModelCore.bias        = sum(oModelCore.Alphas .* Y);
oModelCore.Y           = Y;

% Make predictions on the training set
if(N<1000)
    fClassifierCore        = @(Kext)sum(bsxfun(@times,Kext, oModelCore.Alphas  .* oModelCore.Y))+oModelCore.bias;
    vPred                  = fClassifierCore(Kext(:,:));
    PredictedLabel         = (vPred)<0;
    dAcc                   = sum(PredictedLabel(:) == (Y<0)) / N
else
    vRand   = randsample(N,1000);
    vAlphaY = oModelCore.Alphas.*oModelCore.Y;
    I       = find(oModelCore.Alphas);
    vAlphaY = vAlphaY(I);
    vPred   = zeros(1000,1);
    fClassifierCore        = @(Kext)sum(bsxfun(@times,Kext, vAlphaY))+oModelCore.bias;
    if(numel(I)<10000)        
        vPred                  = fClassifierCore(Kext(I,vRand));
    else
        vBlock=1:min(1000,min(1000,ceil(0.1*1000))):1000;
        for(k=1:numel(vBlock)-1), 
                vPred(vBlock(k):vBlock(k+1)) = (Kext(vRand(vBlock(k):vBlock(k+1)),I)* vAlphaY)+oModelCore.bias;
        end
        vPred(vBlock(end):end)       = (Kext(vRand(vBlock(end):end),I)* vAlphaY)+oModelCore.bias;        
    end
    PredictedLabel         = (vPred)<0;
    dAcc                   = sum(PredictedLabel(:) == (Y(vRand)<0)) / 1000;
end

