% Classification code of indefinite CVM  based on a trained model
% The original training data and the unknown testdata
% Predictions are split into blocks in sample size N is large
% to keep memory usage small
function [cRunResult] = icvm_classify(KOriginal,cModel,mTestData)
    vModels      = cModel.vModels;
    vUniqueL     = cModel.vUniqueL;
    L            = numel(vUniqueL);
    N            = size(mTestData,1);
    vTestIndices = mTestData(:,1);
    vIndicesTrain= cModel.vIndicesTrain;    
    T = cputime;
    cRunResult.Range = 1:N;
    
    DistCenter = zeros(L,N);
    if(N<1000)
        KTestToTrain = KOriginal(vIndicesTrain,vTestIndices)';
        fClassifier  = @(Kext,oModel)sum(bsxfun(@times,Kext, oModel.Alphas  .* oModel.Y))+oModel.bias;    
        for(i=1:L)
            [DistCenter(i,:)] = fClassifier(KTestToTrain',vModels(i));
        end
    else
        vRand            = randsample(N,1000);
        cRunResult.Range = vRand;
        DistCenter       = zeros(L,1000);
        for(i=1:L)
            oModel = vModels(i);
            vAlphaY= oModel.Alphas.*oModel.Y;
            vIndicesRelevant = find(oModel.Alphas);
            vAlphaY = vAlphaY(vIndicesRelevant);
            vPred = zeros(1000,1);
            vBlock=1:min(1000,min(1000,ceil(0.1*1000))):1000;
            for(k=1:numel(vBlock)-1), 
                vPred(vBlock(k):vBlock(k+1)) = (KOriginal(vTestIndices(vRand(vBlock(k):vBlock(k+1))),vIndicesTrain(vIndicesRelevant))* vAlphaY)+oModel.bias;
            end
            vPred(vBlock(end):end)       = (KOriginal(vTestIndices(vRand(vBlock(end):end)),vIndicesTrain(vIndicesRelevant))* vAlphaY)+oModel.bias;                                     
            DistCenter(i,:) = vPred;
            %DistCenter(i,:) = KOriginal(vIndicesTrain,vTestIndices(vRand))' * vAlphaY+oModel.bias;
        end
    end
    Prediction = vUniqueL(argmin(DistCenter)');
    
    dTime = cputime-T;
    cRunResult.Prediction = {Prediction};
    cRunResult.DistCenter = {DistCenter};
    cRunResult.dTime      = dTime;
    
    
